unit armGpio;
{
 ARM GPIO Registry Driver, also can use to manipulate cpu other registry areas

 This code is tested only Samsung s3c2440 cpu, different arm cpus may need different
 gpio driver implementation

 2012 Ermo Kriis and Urmas Tooming
}
{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils;
// cpu datasheet
//http://www.embedinfo.com/english/download/s3c2440.pdf
const
  REG_GPIO = $56000;//s3c2440 gpio register 0x56000000. new fpMap uses page offset, one page is 4096bytes
  // hex 0x1000 so simply calculate 0x56000000 / 0x1000  = 0x56000
  PAGE_SIZE = 4096;

  GPFCON = $50; // IO bank F Config
  GPFDAT = $54; // IO bank F Data
  GPFUP  = $58; // IO bakn F pullup

  // if you need other gpio banks, add here new bank registry offsets
  // and use them: GpIo_Driver.CreatePort(PortConfigOffset, PortDataOffset, PortPullupOffset);
  // NOTE! REG_GPIO defines GpIo start address, each gpio bakn uses own offset NOT real addrss!!
  // example: s3c2440 GpIo register start address is 0x56000, I need to drive bank F so I
  // define offsets GPFCON = $50, GPFDAT = $54; GPFUP = $58;
  // for more information using different gpio banks check cpu datasheet

type
 // predefined IO pin's
  TGpIoPin = (Gp0, Gp1, Gp2, Gp3, Gp4, Gp5, Gp6, Gp7, Gp8, Gp9, Gp10, Gp11, Gp12, Gp13, Gp14, Gp15);

  PDWord = ^DWord;
  PHandle = ^Longint;
  // define all of them so I made Specific1 and Specific2 types. To use them, read cpu datasheet
  TGpioPinConf = (pdInput, pdOutput, pdSpecific1, pdSpecific2);// because each gpio bank has own pin modes I can't

   { TIoPort }

  TIoPort = class // IO bank object, one object can handle up to 16 gpio pin NOTE! s3c2440 GpIo port A is only outputs, and this
  private         // code can't initialise them!!!
   FPortConf: Pointer;
   FPortData: Pointer;
   FPortPullUp: Pointer;
    function get_pinDirection(aPin: TGpIoPin): TGpioPinConf;
    procedure set_PinDirection(aPin: TGpIoPin; AValue: TGpioPinConf);
  public
   property PinConfig[aPin:TGpIoPin]: TGpioPinConf read get_pinDirection write set_PinDirection; // configure selected pin
   function GetBit(const aPin:TGpIoPin):Boolean;inline; // gets pin bit
   procedure ClearBit(const aPin:TGpIoPin);inline;// write pin to 0
   procedure SetBit(const aPin:TGpIoPin);inline;// write pin to 1
   procedure InvertBit(const aPin:TGpIoPin);inline;// invert bit if bit is 0 it will be 1 and 1 will be 0
  end;
    { TIoDriver }

  TIoDriver = class
  private
     FGpIoBaseMap: Pointer;// GpIo memory
     FHandle: PHandle;// pointer of /dev/mem file handle
  public
    destructor Destroy;override;
    function MapIoRegisrty:boolean;// creates io memory mapping
    procedure UnmapIoRegisrty;// close io memory mapping
    function CreatePort(PortConf: Cardinal; PortData: Cardinal; PortUp: Cardinal):TIoPort; // create new IO port
  end;
  //-- ToDo --
  {
    Add other registry drivers
    example
    watchdog
    timers
    power control
  }

  { TMini2440_RegDriver }

  TMini2440_RegDriver = class // mini2440 cpu registry driver,
  private
   FMemHandle: Integer;// /dev/mem file handle
  public
   constructor Create;virtual;reintroduce;
   destructor Destroy; override;
   function OpenDevMem: Boolean;// open memory
   procedure CloseDevMem;// close memory
   function CreateIoDriver: TIoDriver; // create GPIO driver
   // -- ToDo --
   {
    Add functions that creates and maps another areas of cpu registers
    example
    watchdog
    timers
    power control
   }
  end;

implementation

uses
  baseUnix, Unix;

{ TIoDriver }
//*******************************************************************************
destructor TIoDriver.Destroy;
begin
  inherited Destroy;
end;
//*******************************************************************************
function TIoDriver.MapIoRegisrty: boolean;
begin
 Result := True;
 // pascal calls new fpMap, where offst is page sized 4096 bytes!!!
 FGpIoBaseMap := FpMmap(Nil, PAGE_SIZE, PROT_READ or PROT_WRITE, MAP_SHARED, Integer(FHandle), REG_GPIO);
 if Cardinal(FgpIoBaseMap) = 0 then
    Result := False; // unsuccessful memory mapping
end;
//*******************************************************************************
procedure TIoDriver.UnmapIoRegisrty;
begin
  if FGpIoBaseMap <> nil then
 begin
   fpMUnmap(FGpIoBaseMap,PAGE_SIZE);
   FGpIoBaseMap := Nil;
 end;
end;
//*******************************************************************************
function TIoDriver.CreatePort(PortConf: Cardinal; PortData: Cardinal;
  PortUp: Cardinal): TIoPort;
begin
  Result := TIoPort.Create;// new io port
  Result.FPortConf:= Pointer(Cardinal(Self.FGpioBaseMap) + PortConf); // port config memory
  Result.FPortData:= Pointer(Cardinal(Self.FGpioBaseMap) + PortData); // port data memory
  Result.FPortPullUp:= Pointer(Cardinal(Self.FGpioBaseMap) + PortUp); // port pin pullup memory
end;
//*******************************************************************************
{ TMini2440_RegDriver }
constructor TMini2440_RegDriver.Create;
begin
 inherited Create;
 Self.FMemHandle:= 0;
end;
//*******************************************************************************
destructor TMini2440_RegDriver.Destroy;
begin
  Self.CloseDevMem;
  inherited Destroy;
end;
//*******************************************************************************
function TMini2440_RegDriver.OpenDevMem: Boolean;
begin
  Result := True;
try
 Self.FMemHandle:= fpOpen('/dev/mem', O_RDWR or O_SYNC);

 if Self.FMemHandle < 1 then //unsuccessful atemt to open memory file
    Result := False;

except
 Result := False;
end;
end;
//*******************************************************************************
procedure TMini2440_RegDriver.CloseDevMem;
begin
 if FMemHandle > 0 then
    fpClose(FMemHandle);
end;
//*******************************************************************************
function TMini2440_RegDriver.CreateIoDriver: TIoDriver;
begin
 Result := TIODriver.Create;
 Result.FHandle:= PHandle(Self.FMemHandle);
end;
//*******************************************************************************

{ TIoPort }

//NOTE!!!
{
If you need a very fast IO then use following functions directly
because calling another object functions consumes some time.
code like this, gpio speed was 1.8MHz
after direct gpio manipulation, switching speed increases to 3.7MHz
}
function TIoPort.get_pinDirection(aPin: TGpIoPin): TGpioPinConf;
var
  bitL: Byte;
begin
  bitL := Byte(aPin) shl 1;
  result:=TGpioPinConf((PDWord(Self.FPortConf)^ shr  BitL) and 3);
end;
//*******************************************************************************
procedure TIoPort.set_PinDirection(aPin: TGpIoPin; AValue: TGpioPinConf);
var
  bitL, bitH: Byte;
begin
  bitL := Byte(aPin) shl 1;
  bitH := bitL + 1;

  if  (Byte( AValue) and 1) = 0 then
    PDWord(Self.FPortConf)^ := (PDWord(FPortConf)^  and not (1 shl BitL))
  else
    PDWord(Self.FPortConf)^ := (PDWord(FPortConf)^ or (1 shl BitL));

  if  (Byte( AValue) and 2) = 0 then
    PDWord(Self.FPortConf)^ := (PDWord(FPortConf)^  and not (1 shl BitH))
  else
    PDWord(Self.FPortConf)^ := (PDWord(FPortConf)^ or (1 shl BitH));
end;
//*******************************************************************************
function TIoPort.GetBit(const aPin: TGpIoPin): Boolean;
begin
 Result := (PDWord(Self.FPortData)^ and (1 shl Byte(aPin))) <> 0;
end;
//*******************************************************************************
procedure TIoPort.ClearBit(const aPin: TGpIoPin);
begin
 PDWord(FPortData)^ := (PDWord(FPortData)^ and not (1 shl Byte(aPin)));
end;
//*******************************************************************************
procedure TIoPort.SetBit(const aPin: TGpIoPin);
begin
 PDWord(FPortData)^ := (PDWord(FPortData)^ or (1 shl Byte(aPin)));
end;
//*******************************************************************************
procedure TIoPort.InvertBit(const aPin: TGpIoPin);
begin
 PDWord(FPortData)^ := (PDWord(FPortData)^ xor (1 shl Byte(aPin)));
end;
//*******************************************************************************
end.

